/*
 * PostgreSQL functions to interface with memcached using libmemcached.
 * version: 0.2 (2009.01.20)
 *
 * Copyright (c) 2004-2005 Sean Chittenden <sean@chittenden.org>
 * Copyright (c) 2007 Neil Conway <neilc@samurai.com>
 * Copyright (c) 2007 Open Technology Group, Inc. <http://www.otg-nc.com>
 * Copyright (c) 2008-2009 Suzuki Hironobu <hironobu@interdb.jp>
 *
 * See the file COPYING for distribution terms.
 */

#define _NEW_COMMANDS_ 

#include "postgres.h"
#include <inttypes.h>

#include "access/heapam.h"
#include "access/htup.h"
#include "fmgr.h"
#include "funcapi.h"
#include "lib/stringinfo.h"
#include "utils/builtins.h"
#include "utils/datetime.h"
#include "utils/guc.h"
#include "utils/memutils.h"

#include "pgmemcached.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif


/*
 * PostgreSQL <= 8.2 provided VARATT_SIZEP(p), but later versions use
 * SET_VARSIZE() (and also define the VARATT_SIZEP_DEPRECATED
 * symbol). Therefore, define the latter in terms of the former if
 * necessary.
 */
#ifndef SET_VARSIZE
#define SET_VARSIZE(ptr, size) VARATT_SIZEP(ptr) = (size)
#endif

/* Per-backend global state. */
struct memcache_global
{
  memcached_st *mc;
  memcached_server_st *servers;
  memcached_return rc;

  /* context in which long-lived state is allocated */
  MemoryContext pg_ctxt;
};
static struct memcache_global globals;

/* Custom GUC variable */
static char *memcache_default_servers;
static char *memcache_default_behavior;

static const char *assign_default_servers_guc (const char *newval,
					       bool doit, GucSource source);
static const char *assign_default_behavior_guc (const char *newval,
						bool doit, GucSource source);
static const char *assign_default_behavior (const char *newval);
static const char *show_default_servers_guc (void);
static const char *show_default_behavior_guc (void);
static Datum memcache_atomic_op (bool increment, PG_FUNCTION_ARGS);
static Datum memcache_delete_cmd (int type, PG_FUNCTION_ARGS);
static Datum memcache_get_cmd (int type, PG_FUNCTION_ARGS);
static Datum memcache_set_cmd (int type, PG_FUNCTION_ARGS);

static uint64_t get_memcached_distribution_type (const char *data);
static uint64_t get_memcached_hash_type (const char *data);
static memcached_behavior get_memcached_behavior_flag (const char *flag);
static uint64_t get_memcached_behavior_data (const char *flag,
					     const char *data);

static bool do_memcache_set_cmd (int type, char *mkey, size_t mkey_len,
				 char *key, size_t key_len,
				 char *val, size_t val_len, time_t expire);
static time_t interval_to_time_t (Interval * span);

#ifdef _NEW_COMMANDS_
static bool do_server_add(const char *host, const unsigned int port);
#endif

#define MCM_SET_CMD_TYPE_ADD            0x0001
#define MCM_SET_CMD_TYPE_REPLACE        0x0002
#define MCM_SET_CMD_TYPE_SET            0x0004
#define MCM_SET_CMD_TYPE_PREPEND        0x0008
#define MCM_SET_CMD_TYPE_APPEND         0x0010
#define MCM_SET_CMD_TYPE_CAS            0x0020


#define MCM_CMD_TYPE_NORMAL             0x1000
#define MCM_CMD_TYPE_MASTER_KEY         0x2000

#define MCM_BEHAVIOR_SET_TYPE_1         0x0001
#define MCM_BEHAVIOR_SET_TYPE_2         0x0002

void
_PG_init (void)
{
  MemoryContext old_ctxt;

  globals.mc = NULL;
  globals.servers = NULL;
  globals.servers = MEMCACHED_SUCCESS;

  globals.pg_ctxt = AllocSetContextCreate (TopMemoryContext,
					   "pgmemcached global context",
					   ALLOCSET_SMALL_MINSIZE,
					   ALLOCSET_SMALL_INITSIZE,
					   ALLOCSET_SMALL_MAXSIZE);

  old_ctxt = MemoryContextSwitchTo (globals.pg_ctxt);

  globals.mc = memcached_create (NULL);
  if (globals.mc == NULL)
    elog (ERROR, "failed to create libmemcached instance");

  MemoryContextSwitchTo (old_ctxt);

  DefineCustomStringVariable ("pgmemcached.default_servers",
			      "Comma-separated list of memcached servers to connect to.",
			      "Specified as a comma-separated list "
			      "of host:port (port is optional).",
			      &memcache_default_servers,
#if defined(PG_VERSION_NUM) && (80400 <= PG_VERSION_NUM)
			      NULL,
#endif
			      PGC_USERSET,
#if defined(PG_VERSION_NUM) && (80400 <= PG_VERSION_NUM)
                               GUC_LIST_INPUT,
#endif
			      assign_default_servers_guc,
			      show_default_servers_guc);

  DefineCustomStringVariable ("pgmemcached.default_behavior",
			      "Comma-separated list of memcached behavior.",
			      "Specified as a comma-separated list "
			      "of behavior_flag:behavior_data.",
			      &memcache_default_behavior,
#if defined(PG_VERSION_NUM) && (80400 <= PG_VERSION_NUM)
			      NULL,
#endif
			      PGC_USERSET,
#if defined(PG_VERSION_NUM) && (80400 <= PG_VERSION_NUM)
                               GUC_LIST_INPUT,
#endif
			      assign_default_behavior_guc,
			      show_default_behavior_guc);
}


static const char *
assign_default_servers_guc (const char *newval, bool doit, GucSource source)
{
  if ((globals.servers = memcached_servers_parse ((char *) newval)) == NULL)
    return NULL;

  if (memcached_server_push (globals.mc, globals.servers) != MEMCACHED_SUCCESS)
    return NULL;

  return newval;
}


static const char *
assign_default_behavior_guc (const char *newval, bool doit, GucSource source)
{
  return assign_default_behavior (newval);
}

static const char *
assign_default_behavior (const char *newval)
{
  int i;
  int len;
  StringInfoData flag_buf;
  StringInfoData data_buf;

  MemoryContext old_ctx;

  old_ctx = MemoryContextSwitchTo(globals.pg_ctxt);

  initStringInfo (&flag_buf);
  initStringInfo (&data_buf);

  len = strlen (newval);

  for (i = 0; i < len; i++) {
    char c = newval[i];

    if (c == ',' || c == ':') {
      if (flag_buf.len == 0)
	return NULL;

      if (c == ':') {
	int j;

	for (j = i + 1; j < len; j++) {
	  if (newval[j] == ',')
	    break;
	  appendStringInfoChar (&data_buf, newval[j]);
	}

	if (data_buf.len == 0)
	  return NULL;

	i += data_buf.len;
      }

      globals.rc = memcached_behavior_set (globals.mc,
					   get_memcached_behavior_flag
					   (flag_buf.data),
					   get_memcached_behavior_data
					   (flag_buf.data, data_buf.data));

      /* Skip the element separator, reset buffers */
      i++;
      flag_buf.data[0] = '\0';
      flag_buf.len = 0;
      data_buf.data[0] = '\0';
      data_buf.len = 0;
    }
    else {
      appendStringInfoChar (&flag_buf, c);
    }
  }

  MemoryContextSwitchTo(old_ctx);

  pfree (flag_buf.data);
  pfree (data_buf.data);

  return newval;
}


static const char *
show_default_servers_guc (void)
{
  printf ("%s called\n", __FUNCTION__);
  return memcache_default_servers;
}


static const char *
show_default_behavior_guc (void)
{
  printf ("%s called\n", __FUNCTION__);
  return memcache_default_behavior;
}


/*
 * This is called when we're being unloaded from a process. Note that
 * this only happens when we're being replaced by a LOAD (e.g. it
 * doesn't happen on process exit), so we can't depend on it being
 * called.
 */
void
_PG_fini (void)
{
  memcached_server_list_free (globals.servers);
  memcached_free (globals.mc);

  MemoryContextDelete (globals.pg_ctxt);
}


Datum
pgmemcached_set (PG_FUNCTION_ARGS)
{
  return memcache_set_cmd (MCM_SET_CMD_TYPE_SET | MCM_CMD_TYPE_NORMAL,
			   fcinfo);
}


Datum
pgmemcached_set_by_key (PG_FUNCTION_ARGS)
{
  return memcache_set_cmd (MCM_SET_CMD_TYPE_SET | MCM_CMD_TYPE_MASTER_KEY,
			   fcinfo);
}


static Datum
memcache_set_cmd (int type, PG_FUNCTION_ARGS)
{
  text *mkey = NULL;
  text *key;
  text *val;
  size_t mkey_len = 0;
  size_t key_len;
  size_t val_len;
  time_t expire = 0.0;
  bool ret;
  int arg = 0;

  if (type & MCM_CMD_TYPE_MASTER_KEY) {
    if (PG_ARGISNULL (arg))
      ereport (ERROR,
	       (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
		errmsg ("memcache master key cannot be NULL")));

    mkey = PG_GETARG_TEXT_P (arg);
    mkey_len = VARSIZE (mkey) - VARHDRSZ;
    if (mkey_len < 1)
      ereport (ERROR,
	       (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
		errmsg ("memcache master key cannot be the empty string")));
    arg++;
  }

  if (PG_ARGISNULL (arg))
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcache key cannot be NULL")));

  key = PG_GETARG_TEXT_P (arg);
  key_len = VARSIZE (key) - VARHDRSZ;
  if (key_len < 1)
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcache key cannot be the empty string")));
  arg++;

  if (PG_ARGISNULL (arg))
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcache value cannot be NULL")));
  val = PG_GETARG_TEXT_P (arg);
  val_len = VARSIZE (val) - VARHDRSZ;
  arg++;

  if (PG_NARGS () >= (arg + 1) && PG_ARGISNULL (arg) == false) {
    Interval *span = PG_GETARG_INTERVAL_P (arg);
    expire = interval_to_time_t (span);
    arg++;
  }

  ret =
    do_memcache_set_cmd (type, VARDATA (mkey), mkey_len, VARDATA (key),
			 key_len, VARDATA (val), val_len, expire);

  PG_RETURN_BOOL (ret);
}


static bool
do_memcache_set_cmd (int type, char *mkey, size_t mkey_len, char *key,
		     size_t key_len, char *val, size_t val_len, time_t expire)
{
  if (type & MCM_SET_CMD_TYPE_ADD)
    if (type & MCM_CMD_TYPE_MASTER_KEY)
      globals.rc =
	memcached_add_by_key (globals.mc, mkey, mkey_len, key, key_len, val,
			      val_len, expire, 0);
    else
      globals.rc =
	memcached_add (globals.mc, key, key_len, val, val_len, expire, 0);
  else if (type & MCM_SET_CMD_TYPE_REPLACE)
    if (type & MCM_CMD_TYPE_MASTER_KEY)
      globals.rc =
	memcached_replace_by_key (globals.mc, mkey, mkey_len, key, key_len,
				  val, val_len, expire, 0);
    else
      globals.rc =
	memcached_replace (globals.mc, key, key_len, val, val_len, expire, 0);
  else if (type & MCM_SET_CMD_TYPE_SET)
    if (type & MCM_CMD_TYPE_MASTER_KEY)
      globals.rc =
	memcached_set_by_key (globals.mc, mkey, mkey_len, key, key_len, val,
			      val_len, expire, 0);
    else
      globals.rc =
	memcached_set (globals.mc, key, key_len, val, val_len, expire, 0);
  else if (type & MCM_SET_CMD_TYPE_PREPEND)
    if (type & MCM_CMD_TYPE_MASTER_KEY)
      globals.rc =
	memcached_prepend_by_key (globals.mc, mkey, mkey_len, key, key_len,
				  val, val_len, expire, 0);
    else
      globals.rc =
	memcached_prepend (globals.mc, key, key_len, val, val_len, expire, 0);
  else if (type & MCM_SET_CMD_TYPE_APPEND)
    if (type & MCM_CMD_TYPE_MASTER_KEY)
      globals.rc =
	memcached_append_by_key (globals.mc, mkey, mkey_len, key, key_len,
				 val, val_len, expire, 0);
    else
      globals.rc =
	memcached_append (globals.mc, key, key_len, val, val_len, expire, 0);
  else
    elog (ERROR, "unknown memcached set command type: %d", type);

  if (globals.rc != MEMCACHED_SUCCESS) {
    elog (ERROR, "memcached_result: %s", memcached_strerror(globals.mc, globals.rc));
    return (false);
  }

  return (true);
}


Datum
pgmemcached_add (PG_FUNCTION_ARGS)
{
  return memcache_set_cmd (MCM_SET_CMD_TYPE_ADD | MCM_CMD_TYPE_NORMAL,
			   fcinfo);
}


Datum
pgmemcached_add_by_key (PG_FUNCTION_ARGS)
{
  return memcache_set_cmd (MCM_SET_CMD_TYPE_ADD | MCM_CMD_TYPE_MASTER_KEY,
			   fcinfo);
}


Datum
pgmemcached_replace (PG_FUNCTION_ARGS)
{
  return memcache_set_cmd (MCM_SET_CMD_TYPE_REPLACE | MCM_CMD_TYPE_NORMAL,
			   fcinfo);
}


Datum
pgmemcached_replace_by_key (PG_FUNCTION_ARGS)
{
  return memcache_set_cmd (MCM_SET_CMD_TYPE_REPLACE | MCM_CMD_TYPE_MASTER_KEY,
			   fcinfo);
}


Datum
pgmemcached_prepend (PG_FUNCTION_ARGS)
{
  return memcache_set_cmd (MCM_SET_CMD_TYPE_PREPEND | MCM_CMD_TYPE_NORMAL,
			   fcinfo);
}


Datum
pgmemcached_prepend_by_key (PG_FUNCTION_ARGS)
{
  return memcache_set_cmd (MCM_SET_CMD_TYPE_PREPEND | MCM_CMD_TYPE_MASTER_KEY,
			   fcinfo);
}


Datum
pgmemcached_append (PG_FUNCTION_ARGS)
{
  return memcache_set_cmd (MCM_SET_CMD_TYPE_APPEND | MCM_CMD_TYPE_NORMAL,
			   fcinfo);
}


Datum
pgmemcached_append_by_key (PG_FUNCTION_ARGS)
{
  return memcache_set_cmd (MCM_SET_CMD_TYPE_APPEND | MCM_CMD_TYPE_MASTER_KEY,
			   fcinfo);
}


static Datum
memcache_atomic_op (bool increment, PG_FUNCTION_ARGS)
{
  text *key = PG_GETARG_TEXT_P (0);
  size_t key_len;
  u_int64_t val = 0;
  unsigned int offset = 1;

  key_len = VARSIZE (key) - VARHDRSZ;
  if (key_len < 1)
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("unable to have a zero length key")));

  if (PG_NARGS () >= 2)
    offset = PG_GETARG_UINT32 (1);

  if (increment)
    globals.rc =
      memcached_increment (globals.mc, VARDATA (key), key_len, offset, &val);
  else
    globals.rc =
      memcached_decrement (globals.mc, VARDATA (key), key_len, offset, &val);

  if (globals.rc != MEMCACHED_SUCCESS) {
    elog (ERROR, "memcached_result: %s", memcached_strerror(globals.mc, globals.rc));
    PG_RETURN_NULL ();
  }

  PG_RETURN_INT32 (val);
}


Datum
pgmemcached_incr (PG_FUNCTION_ARGS)
{
  return memcache_atomic_op (true, fcinfo);
}


Datum
pgmemcached_decr (PG_FUNCTION_ARGS)
{
  return memcache_atomic_op (false, fcinfo);
}


static Datum
memcache_get_cmd (int type, PG_FUNCTION_ARGS)
{
  text *mkey = NULL;
  text *key;
  text *ret;
  char *data;
  size_t mkey_len = 0;
  size_t key_len;
  size_t value_len;
  unsigned int flag;
  memcached_return rc;
  int arg = 0;

  if (type & MCM_CMD_TYPE_MASTER_KEY) {
    if (PG_ARGISNULL (arg))
      ereport (ERROR,
	       (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
		errmsg ("memcache master key cannot be NULL")));

    mkey = PG_GETARG_TEXT_P (arg);
    mkey_len = VARSIZE (mkey) - VARHDRSZ;
    if (mkey_len < 1)
      ereport (ERROR,
	       (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
		errmsg ("memcache master key cannot be the empty string")));
    arg++;
  }

  if (PG_ARGISNULL (arg))
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcache key cannot be NULL")));

  key = PG_GETARG_TEXT_P (arg);
  key_len = VARSIZE (key) - VARHDRSZ;
  if (key_len < 1)
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcache key cannot be the empty string")));

  if (type & MCM_CMD_TYPE_MASTER_KEY)
    data =
      memcached_get_by_key (globals.mc, VARDATA (mkey), mkey_len,
			    VARDATA (key), key_len, &value_len, &flag, &rc);
  else
    data = memcached_get (globals.mc, VARDATA (key), key_len,
			  &value_len, &flag, &rc);

  globals.rc = rc;
  if (rc != MEMCACHED_SUCCESS) {
    elog (ERROR, "memcached_result: %s", memcached_strerror(globals.mc, globals.rc));
    PG_RETURN_TEXT_P (NULL);
  }

  ret = (text *) palloc (value_len + VARHDRSZ);
  SET_VARSIZE (ret, value_len + VARHDRSZ);
  memcpy (VARDATA (ret), data, value_len);

  PG_RETURN_TEXT_P (ret);
}


Datum
pgmemcached_get (PG_FUNCTION_ARGS)
{
  return memcache_get_cmd (MCM_CMD_TYPE_NORMAL, fcinfo);
}


Datum
pgmemcached_get_by_key (PG_FUNCTION_ARGS)
{
  return memcache_get_cmd (MCM_CMD_TYPE_MASTER_KEY, fcinfo);
}


static Datum
memcache_delete_cmd (int type, PG_FUNCTION_ARGS)
{
  text *mkey = NULL;
  size_t mkey_len = 0;
  text *key;
  size_t key_len;
  time_t hold;
  int arg = 0;

  if (type & MCM_CMD_TYPE_MASTER_KEY) {
    mkey = PG_GETARG_TEXT_P (arg++);
    mkey_len = VARSIZE (mkey) - VARHDRSZ;
    if (mkey_len < 1)
      ereport (ERROR,
	       (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
		errmsg ("memcached master_key cannot be the empty string")));
  }

  key = PG_GETARG_TEXT_P (arg++);
  key_len = VARSIZE (key) - VARHDRSZ;
  if (key_len < 1)
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcached key cannot be the empty string")));

  hold = (time_t) 0.0;
  if (PG_NARGS () >= (arg + 1) && PG_ARGISNULL (arg) == false)
    hold = interval_to_time_t (PG_GETARG_INTERVAL_P (arg));

  if (type & MCM_CMD_TYPE_MASTER_KEY)
    globals.rc = memcached_delete_by_key (globals.mc, VARDATA (mkey), mkey_len,
					  VARDATA (key), key_len, hold);
  else
    globals.rc = memcached_delete (globals.mc, VARDATA (key), key_len, hold);
  
  if (globals.rc != MEMCACHED_SUCCESS) {
    elog (ERROR, "memcached_result: %s", memcached_strerror(globals.mc, globals.rc));
    PG_RETURN_BOOL (false);
  }

  PG_RETURN_BOOL (true);
}


Datum
pgmemcached_delete (PG_FUNCTION_ARGS)
{
  return memcache_delete_cmd (MCM_CMD_TYPE_NORMAL, fcinfo);
}


Datum
pgmemcached_delete_by_key (PG_FUNCTION_ARGS)
{
  return memcache_delete_cmd (MCM_CMD_TYPE_MASTER_KEY, fcinfo);
}


Datum
pgmemcached_flush (PG_FUNCTION_ARGS)
{
  time_t expire = 0.0;

  if (PG_NARGS () >= 1 && PG_ARGISNULL (0) == false) {
    Interval *span = PG_GETARG_INTERVAL_P (0);
    expire = interval_to_time_t (span);
  }

  if ((globals.rc =
       memcached_flush (globals.mc, expire)) != MEMCACHED_SUCCESS) {
    elog (ERROR, "memcached_result: %s", memcached_strerror(globals.mc, globals.rc));
    PG_RETURN_BOOL (false);
  }
  PG_RETURN_BOOL (true);
}


Datum
pgmemcached_return (PG_FUNCTION_ARGS)
{
  PG_RETURN_INT32 (globals.rc);
}


Datum
pgmemcached_strerror (PG_FUNCTION_ARGS)
{
  memcached_return rc = globals.rc;
  text *ret;
  char *msg = NULL;
  size_t msg_len;

  if (PG_NARGS () == 1)
    rc = (memcached_return) PG_GETARG_INT32 (0);

  msg = memcached_strerror (globals.mc, rc);
  msg_len = strlen (msg);
  ret = (text *) palloc (msg_len + VARHDRSZ);
  SET_VARSIZE (ret, msg_len + VARHDRSZ);
  memcpy (VARDATA (ret), msg, msg_len);

  PG_RETURN_TEXT_P (ret);
}


Datum
pgmemcached_quit (PG_FUNCTION_ARGS)
{
  memcached_quit (globals.mc);
  globals.rc = MEMCACHED_SUCCESS;
  PG_RETURN_BOOL (true);
}


static time_t
interval_to_time_t (Interval * span)
{
  float8 result;
  result = span->time / 1000000e0;

  if (span->month != 0) {
    result += (365.25 * 86400) * (span->month / 12);
    result += (30.0 * 86400) * (span->month % 12);
  }

  return (time_t) result;
}


static memcached_behavior
get_memcached_behavior_flag (const char *flag)
{
  memcached_behavior ret = -1;

  /*
   * Sort by flag in reverse order.
   */
  if (strncmp ("MEMCACHED_BEHAVIOR_HASH_WITH_PREFIX_KEY", flag, 39) == 0
	   || strncmp ("HASH_WITH_PREFIX_KEY", flag, 21) == 0)
    ret = MEMCACHED_BEHAVIOR_HASH_WITH_PREFIX_KEY;
  else if (strncmp ("MEMCACHED_BEHAVIOR_VERIFY_KEY", flag, 29) == 0
	   || strncmp ("VERIFY_KEY", flag, 10) == 0)
    ret = MEMCACHED_BEHAVIOR_VERIFY_KEY;
  else if (strncmp ("MEMCACHED_BEHAVIOR_USER_DATA", flag, 28) == 0
	   || strncmp ("USER_DATA", flag, 9) == 0)
    ret = MEMCACHED_BEHAVIOR_USER_DATA;
  else if (strncmp ("MEMCACHED_BEHAVIOR_TCP_NODELAY", flag, 30) == 0
	   || strncmp ("TCP_NODELAY", flag, 11) == 0)
    ret = MEMCACHED_BEHAVIOR_TCP_NODELAY;
  else if (strncmp ("MEMCACHED_BEHAVIOR_SUPPORT_CAS", flag, 30) == 0
	   || strncmp ("SUPPORT_CAS", flag, 11) == 0)
    ret = MEMCACHED_BEHAVIOR_SUPPORT_CAS;
  else if (strncmp ("MEMCACHED_BEHAVIOR_SORT_HOSTS", flag, 29) == 0
	   || strncmp ("SORT_HOSTS", flag, 10) == 0)
    ret = MEMCACHED_BEHAVIOR_SORT_HOSTS;
  else if (strncmp ("MEMCACHED_BEHAVIOR_SOCKET_SEND_SIZE", flag, 35) == 0
	   || strncmp ("SOCKET_SEND_SIZE", flag, 16) == 0)
    ret = MEMCACHED_BEHAVIOR_SOCKET_SEND_SIZE;
  else if (strncmp ("MEMCACHED_BEHAVIOR_SOCKET_RECV_SIZE", flag, 35) == 0
	   || strncmp ("SOCKET_RECV_SIZE", flag, 16) == 0)
    ret = MEMCACHED_BEHAVIOR_SOCKET_RECV_SIZE;
  else if (strncmp ("MEMCACHED_BEHAVIOR_SND_TIMEOUT", flag, 30) == 0
	   || strncmp ("SND_TIMEOUT", flag, 11) == 0)
    ret = MEMCACHED_BEHAVIOR_SND_TIMEOUT;
  else if (strncmp ("MEMCACHED_BEHAVIOR_SERVER_FAILURE_LIMIT", flag, 39) == 0
	   || strncmp ("SERVER_FAILURE_LIMIT", flag, 20) == 0)
    ret = MEMCACHED_BEHAVIOR_SERVER_FAILURE_LIMIT;
  else if (strncmp ("MEMCACHED_BEHAVIOR_RETRY_TIMEOUT", flag, 32) == 0
	   || strncmp ("RETRY_TIMEOUT", flag, 13) == 0)
    ret = MEMCACHED_BEHAVIOR_RETRY_TIMEOUT;
  else if (strncmp ("MEMCACHED_BEHAVIOR_RCV_TIMEOUT", flag, 30) == 0
	   || strncmp ("RCV_TIMEOUT", flag, 11) == 0)
    ret = MEMCACHED_BEHAVIOR_RCV_TIMEOUT;
  else if (strncmp ("MEMCACHED_BEHAVIOR_POLL_TIMEOUT", flag, 31) == 0
	   || strncmp ("POLL_TIMEOUT", flag, 12) == 0)
    ret = MEMCACHED_BEHAVIOR_POLL_TIMEOUT;
  else if (strncmp ("MEMCACHED_BEHAVIOR_NO_BLOCK", flag, 27) == 0
      || strncmp ("NO_BLOCK", flag, 8) == 0)
    ret = MEMCACHED_BEHAVIOR_NO_BLOCK;
  else if (strncmp ("MEMCACHED_BEHAVIOR_KETAMA_WEIGHTED", flag, 34) == 0
	   || strncmp ("KETAMA_WEIGHTED", flag, 15) == 0)
    ret = MEMCACHED_BEHAVIOR_KETAMA_WEIGHTED;
  else if (strncmp ("MEMCACHED_BEHAVIOR_KETAMA_HASH", flag, 30) == 0
	   || strncmp ("KETAMA_HASH", flag, 11) == 0)
    ret = MEMCACHED_BEHAVIOR_KETAMA_HASH;
  else if (strncmp ("MEMCACHED_BEHAVIOR_KETAMA", flag, 25) == 0
	   || strncmp ("KETAMA", flag, 6) == 0)
    ret = MEMCACHED_BEHAVIOR_KETAMA;
  else if (strncmp ("MEMCACHED_BEHAVIOR_IO_MSG_WATERMARK", flag, 35) == 0
	   || strncmp ("IO_MSG_WATERMARK", flag, 16) == 0)
    ret = MEMCACHED_BEHAVIOR_IO_MSG_WATERMARK;
  else if (strncmp ("MEMCACHED_BEHAVIOR_IO_BYTES_WATERMARK", flag, 37) == 0
	   || strncmp ("IO_BYTES_WATERMARK", flag, 18) == 0)
    ret = MEMCACHED_BEHAVIOR_IO_BYTES_WATERMARK;
  else if (strncmp ("MEMCACHED_BEHAVIOR_HASH", flag, 23) == 0
	   || strncmp ("HASH", flag, 4) == 0)
    ret = MEMCACHED_BEHAVIOR_HASH;
  else if (strncmp ("MEMCACHED_BEHAVIOR_DISTRIBUTION", flag, 31) == 0
	   || strncmp ("DISTRIBUTION", flag, 12) == 0)
    ret = MEMCACHED_BEHAVIOR_DISTRIBUTION;
  else if (strncmp ("MEMCACHED_BEHAVIOR_CONNECT_TIMEOUT", flag, 34) == 0
	   || strncmp ("CONNECT_TIMEOUT", flag, 15) == 0)
    ret = MEMCACHED_BEHAVIOR_CONNECT_TIMEOUT;
  else if (strncmp ("MEMCACHED_BEHAVIOR_CACHE_LOOKUPS", flag, 33) == 0
	   || strncmp ("CACHE_LOOKUPS", flag, 14) == 0)
    ret = MEMCACHED_BEHAVIOR_CACHE_LOOKUPS;
  else if (strncmp ("MEMCACHED_BEHAVIOR_BUFFER_REQUESTS", flag, 34) == 0
	   || strncmp ("BUFFER_REQUESTS", flag, 15) == 0)
    ret = MEMCACHED_BEHAVIOR_BUFFER_REQUESTS;
  else if (strncmp ("MEMCACHED_BEHAVIOR_BINARY_PROTOCOL", flag, 34) == 0
	   || strncmp ("BINARY_PROTOCOL", flag, 15) == 0)
    ret = MEMCACHED_BEHAVIOR_BINARY_PROTOCOL;
  else
    elog (ERROR, "unknown memcached behavior flag: %s", flag);

  return ret;
}


static uint64_t
get_memcached_hash_type (const char *data)
{
  uint64_t ret;

  /*
   * Sort by data in reverse order.
   */
  if (strncmp ("MEMCACHED_HASH_MURMUR", data, 21) == 0
	   || strncmp ("MURMUR", data, 6) == 0)
    ret = MEMCACHED_HASH_MURMUR;
  else if (strncmp ("MEMCACHED_HASH_MD5", data, 18) == 0
	   || strncmp ("MD5", data, 3) == 0)
    ret = MEMCACHED_HASH_MD5;
  else if (strncmp ("MEMCACHED_HASH_JENKINS", data, 22) == 0
	   || strncmp ("JENKINS", data, 7) == 0)
    ret = MEMCACHED_HASH_JENKINS;
  else if (strncmp ("MEMCACHED_HASH_HSIEH", data, 20) == 0
	   || strncmp ("HSIEH", data, 5) == 0)
    ret = MEMCACHED_HASH_HSIEH;
  else if (strncmp ("MEMCACHED_HASH_FNV1A_64", data, 23) == 0
	   || strncmp ("FNV1A_64", data, 8) == 0)
    ret = MEMCACHED_HASH_FNV1A_64;
  else if (strncmp ("MEMCACHED_HASH_FNV1A_32", data, 23) == 0
	   || strncmp ("FNV1A_32", data, 8) == 0)
    ret = MEMCACHED_HASH_FNV1A_32;
  else if (strncmp ("MEMCACHED_HASH_FNV1_64", data, 22) == 0
	   || strncmp ("FNV1_64", data, 7) == 0)
    ret = MEMCACHED_HASH_FNV1_64;
  else if (strncmp ("MEMCACHED_HASH_FNV1_32", data, 22) == 0
	   || strncmp ("FNV1_32", data, 7) == 0)
    ret = MEMCACHED_HASH_FNV1_32;
  else if (strncmp ("MEMCACHED_HASH_DEFAULT", data, 22) == 0
      || strncmp ("DEFAULT", data, 7) == 0)
    ret = MEMCACHED_HASH_DEFAULT;
  else if (strncmp ("MEMCACHED_HASH_CRC", data, 18) == 0
	   || strncmp ("CRC", data, 3) == 0)
    ret = MEMCACHED_HASH_CRC;
  else {
    ret = 0xffffffff; /* to avoid warning */
    elog (ERROR, "invalid hash name: %s", data);
  }
  return ret;
}



static uint64_t
get_memcached_distribution_type (const char *data)
{
  uint64_t ret;

  /*
   * Sort by data in reverse order.
   */
  if (strncmp ("MEMCACHED_DISTRIBUTION_RANDOM", data, 29) == 0
	   || strncmp ("RANDOM", data, 6) == 0)
    ret = MEMCACHED_DISTRIBUTION_RANDOM;
  else if (strncmp ("MEMCACHED_DISTRIBUTION_MODULA", data, 29) == 0
      || strncmp ("MODULA", data, 6) == 0)
    ret = MEMCACHED_DISTRIBUTION_MODULA;
  else if (strncmp ("MEMCACHED_DISTRIBUTION_CONSISTENT_KETAMA", data, 40) == 0
	   || strncmp ("CONSISTENT_KETAMA", data, 17) == 0)
    ret = MEMCACHED_DISTRIBUTION_CONSISTENT_KETAMA;
  else if (strncmp ("MEMCACHED_DISTRIBUTION_CONSISTENT", data, 33) == 0
	   || strncmp ("CONSISTENT", data, 10) == 0)
    ret = MEMCACHED_DISTRIBUTION_CONSISTENT;
  else {
    ret = 0xffffffff; /* to avoid warning */
    elog (ERROR, "invalid distribution name: %s", data);
  }
  return ret;
}


static uint64_t
get_memcached_behavior_data (const char *flag, const char *data)
{
  char *endptr;
  memcached_behavior f_code = get_memcached_behavior_flag (flag);
  uint64_t ret;

  switch (f_code) {
  case MEMCACHED_BEHAVIOR_HASH:
  case MEMCACHED_BEHAVIOR_KETAMA_HASH:
    ret = get_memcached_hash_type (data);
    break;
  case MEMCACHED_BEHAVIOR_DISTRIBUTION:
    ret = get_memcached_distribution_type (data);
    break;
  default:
    ret = strtol (data, &endptr, 10);
    if (endptr == data)
      elog (ERROR, "invalid memcached behavior param %s: %s", flag, data);
  }
  return ret;
}

static bool
do_memcached_behavior_set (int type, PG_FUNCTION_ARGS)
{
  text *flag;
  text *data;
  char *flag_str;
  char *data_str;
  size_t flag_len;
  size_t data_len;
  memcached_behavior flag_code;
  uint64_t data_code;

  if (PG_ARGISNULL (0))
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcache behavior flag cannot be NULL")));

  flag = PG_GETARG_TEXT_P (0);
  flag_len = VARSIZE (flag) - VARHDRSZ;
  if (flag_len < 1)
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcache behavior flag cannot be the empty string")));

  flag_str = DatumGetCString (DirectFunctionCall1 (textout,
						   PointerGetDatum (flag)));

  if ((flag_code = get_memcached_behavior_flag (flag_str)) == -1)
    return (false);

  if (PG_ARGISNULL (1))
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcache behavior data cannot be NULL")));

  if (type == MCM_BEHAVIOR_SET_TYPE_1) {
    data = PG_GETARG_TEXT_P (1);
    data_len = VARSIZE (data) - VARHDRSZ;
    if (data_len < 1)
      ereport (ERROR,
	       (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
		errmsg
		("memcache behavior data cannot be the empty string")));

    data_str = DatumGetCString (DirectFunctionCall1 (textout,
						     PointerGetDatum (data)));
    data_code = get_memcached_behavior_data (flag_str, data_str);
  }
  else {
    data_code = (uint64_t) PG_GETARG_INT64 (1);
  }

  globals.rc = memcached_behavior_set (globals.mc, flag_code, data_code);

  if (globals.rc != MEMCACHED_SUCCESS) {
    elog (ERROR, "memcached_result: %s", memcached_strerror(globals.mc, globals.rc));
    return (false);
  }

  return (true);
}


Datum
pgmemcached_behavior_set1 (PG_FUNCTION_ARGS)
{
  PG_RETURN_BOOL (do_memcached_behavior_set
		  (MCM_BEHAVIOR_SET_TYPE_1, fcinfo));
}


Datum
pgmemcached_behavior_set2 (PG_FUNCTION_ARGS)
{
  PG_RETURN_BOOL (do_memcached_behavior_set
		  (MCM_BEHAVIOR_SET_TYPE_2, fcinfo));
}


Datum
pgmemcached_behavior_get (PG_FUNCTION_ARGS)
{
  text *flag;
  char *flag_str;
  size_t flag_len;
  memcached_behavior flag_code;
  uint64_t ret;

  if (PG_ARGISNULL (0))
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcache behavior flag cannot be NULL")));

  flag = PG_GETARG_TEXT_P (0);
  flag_len = VARSIZE (flag) - VARHDRSZ;
  if (flag_len < 1)
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcache behavior flag cannot be the empty string")));

  flag_str = DatumGetCString (DirectFunctionCall1 (textout,
						   PointerGetDatum (flag)));

  if ((flag_code = get_memcached_behavior_flag (flag_str)) == -1)
    PG_RETURN_NULL ();

  ret = memcached_behavior_get (globals.mc, flag_code);
  globals.rc = MEMCACHED_SUCCESS;

  PG_RETURN_INT64 (ret);	// TODO 64 --> u64
}


#ifdef _NEW_COMMANDS_
Datum
pgmemcached_create (PG_FUNCTION_ARGS)
{
  MemoryContext old_ctxt;

  old_ctxt = MemoryContextSwitchTo (globals.pg_ctxt);

  if (globals.mc != NULL)
    memcached_free (globals.mc);

  globals.mc = memcached_create (NULL);

  MemoryContextSwitchTo (old_ctxt);
  if (globals.mc == NULL) {
    elog (ERROR, "failed to create libmemcached instance");
    globals.rc = MEMCACHED_FAILURE;
    PG_RETURN_BOOL (false);
  }
  globals.rc = MEMCACHED_SUCCESS;
  PG_RETURN_BOOL (true);
}


Datum
pgmemcached_free (PG_FUNCTION_ARGS)
{
  MemoryContext old_ctxt;

  old_ctxt = MemoryContextSwitchTo (globals.pg_ctxt);
  if (globals.mc != NULL)
    memcached_free (globals.mc);
  globals.mc = NULL;
  globals.rc = MEMCACHED_SUCCESS;
  MemoryContextSwitchTo (old_ctxt);

  PG_RETURN_BOOL (true);
}


Datum
pgmemcached_server_count (PG_FUNCTION_ARGS)
{
  unsigned int count = 0;

  if (globals.mc != NULL)
    count = memcached_server_count (globals.mc);
  globals.rc = MEMCACHED_SUCCESS;
  PG_RETURN_UINT32 (count);
}


Datum
pgmemcached_server_list (PG_FUNCTION_ARGS)
{
  MemoryContext old_ctxt;
  bool ret = true;

  old_ctxt = MemoryContextSwitchTo (globals.pg_ctxt);
  globals.rc = MEMCACHED_FAILURE;
  if (globals.servers != NULL)
    ret = false;
  else if ((globals.servers = memcached_server_list (globals.mc)) == NULL)
    ret = false;
  globals.rc = MEMCACHED_SUCCESS;
  MemoryContextSwitchTo (old_ctxt);

  PG_RETURN_BOOL (ret);
}


Datum
pgmemcached_server_list_free (PG_FUNCTION_ARGS)
{
  MemoryContext old_ctxt;

  old_ctxt = MemoryContextSwitchTo (globals.pg_ctxt);
  if (globals.servers != NULL)
    memcached_server_list_free (globals.servers);
  globals.servers = NULL;
  globals.rc = MEMCACHED_SUCCESS;
  MemoryContextSwitchTo (old_ctxt);

  PG_RETURN_BOOL (true);
}


Datum
pgmemcached_server_list_count (PG_FUNCTION_ARGS)
{
  unsigned int count = 0;

  if (globals.servers != NULL)
    count = memcached_server_list_count (globals.servers);
  globals.rc = MEMCACHED_SUCCESS;
  PG_RETURN_UINT32 (count);
}


Datum
pgmemcached_server_display (PG_FUNCTION_ARGS)
{
  unsigned int count = 0;
  FuncCallContext *func_ctx;
  int call_cntr;
  int max_calls;
  AttInMetadata *attinmeta;
  memcached_server_st *list;

  if (globals.mc != NULL)
    count = memcached_server_count (globals.mc);
  else
    PG_RETURN_NULL ();

  if (SRF_IS_FIRSTCALL ()) {
    MemoryContext old_ctx;
    TypeFuncClass func_type;
    TupleDesc result_tdesc;

    func_ctx = SRF_FIRSTCALL_INIT ();

    old_ctx = MemoryContextSwitchTo (func_ctx->multi_call_memory_ctx);

    func_type = get_call_result_type (fcinfo, NULL, &result_tdesc);
    if (func_type != TYPEFUNC_COMPOSITE)
      elog (ERROR, "called in unexpected context");

    func_ctx->max_calls = count;
    attinmeta = TupleDescGetAttInMetadata (result_tdesc);
    func_ctx->attinmeta = attinmeta;

    MemoryContextSwitchTo (old_ctx);
  }

  func_ctx = SRF_PERCALL_SETUP ();

  call_cntr = func_ctx->call_cntr;
  max_calls = func_ctx->max_calls;
  attinmeta = func_ctx->attinmeta;
  list = globals.mc->hosts;

  if (call_cntr < max_calls)  {
    char **values;
    HeapTuple tuple;
    Datum result;
    
    values = (char **) palloc (2 * sizeof (char *));
    values[0] = (char *) palloc (MEMCACHED_MAX_HOST_LENGTH * sizeof (char));
    values[1] = (char *) palloc (16 * sizeof (char));
    
    snprintf (values[0], strlen (list[call_cntr].hostname) + 1, "%s",
	      list[call_cntr].hostname);
    snprintf (values[1], 16, "%d", list[call_cntr].port);
    
    tuple = BuildTupleFromCStrings (attinmeta, values);
    result = HeapTupleGetDatum (tuple);
    
    pfree (values[0]);
    pfree (values[1]);
    pfree (values);
    
    SRF_RETURN_NEXT (func_ctx, result);
  }
  else    {  
    SRF_RETURN_DONE (func_ctx);
  }
}


Datum
pgmemcached_server_list_append (PG_FUNCTION_ARGS)
{
  text *host;
  char *host_str;
  size_t host_len;
  unsigned int port;
  memcached_return rc;
  MemoryContext old_ctxt;

  old_ctxt = MemoryContextSwitchTo (globals.pg_ctxt);

  if (PG_ARGISNULL (0))
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcache server list append host cannot be NULL")));
  host = PG_GETARG_TEXT_P (0);
  host_len = VARSIZE (host) - VARHDRSZ;
  if (host_len < 1)
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg
	      ("memcache server list append host cannot be the empty string")));

  host_str = DatumGetCString (DirectFunctionCall1 (textout,
						   PointerGetDatum (host)));

  if (PG_ARGISNULL (1))
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcache server list append port cannot be NULL")));

  port = (unsigned int) PG_GETARG_UINT32 (1);

  globals.servers =
    memcached_server_list_append (globals.servers, host_str, port, &rc);

  MemoryContextSwitchTo (old_ctxt);

  globals.rc = rc;
  if (rc != MEMCACHED_SUCCESS) {
    elog (ERROR, "memcached_result: %s", memcached_strerror(globals.mc, globals.rc));
    PG_RETURN_BOOL (false);
  }

  PG_RETURN_BOOL (true);
}


Datum
pgmemcached_server_push (PG_FUNCTION_ARGS)
{
  MemoryContext old_ctxt;
  bool ret = true;

  old_ctxt = MemoryContextSwitchTo (globals.pg_ctxt);

  if (globals.mc == NULL || globals.servers == NULL)
    ret = false;
  else if ((globals.rc = memcached_server_push (globals.mc, globals.servers))
	   != MEMCACHED_SUCCESS) {
    elog (ERROR, "memcached_result: %s", memcached_strerror(globals.mc, globals.rc));
    ret = false;
  }

  if (memcache_default_behavior != NULL)
    assign_default_behavior (memcache_default_behavior);

  MemoryContextSwitchTo (old_ctxt);

  PG_RETURN_BOOL (ret);
}


static bool
do_server_add(const char *host, const unsigned int port)
{
  MemoryContext old_ctx;

  /* Ensure that new servers are allocated in the long-term context */
  old_ctx = MemoryContextSwitchTo(globals.pg_ctxt);

  globals.rc = memcached_server_add (globals.mc, (char *)host, (unsigned int)port);

  if (globals.rc != MEMCACHED_SUCCESS) {
    MemoryContextSwitchTo(old_ctx);

    elog (ERROR, "memcached_result: %s", memcached_strerror(globals.mc, globals.rc));
    return false;
  }

  MemoryContextSwitchTo(old_ctx);
  return true;
}

Datum
pgmemcached_server_add(PG_FUNCTION_ARGS)
{
  text          *server = PG_GETARG_TEXT_P(0);
  char          *server_data;
  /* If no port specified, use the libmemcache default (11211) */
  unsigned int  port = 11211;

  server_data = DatumGetCString(DirectFunctionCall1(textout,
						    PointerGetDatum(server)));

  if (PG_NARGS() > 1)
    port = (unsigned int) PG_GETARG_UINT32 (1);

  if (do_server_add(server_data, port) == false)
    {
      elog(NOTICE, "unable to add memcached server");
      PG_RETURN_BOOL(false);
    }

  PG_RETURN_BOOL(true);
}

#endif


Datum
pgmemcached_stat (PG_FUNCTION_ARGS)
{
  memcached_stat_st *stats;
  memcached_return rc;
  StringInfoData buf;

  stats = memcached_stat (globals.mc, NULL, &rc);
  globals.rc = rc;

  if (stats == NULL)
    elog (ERROR, "unable to create a new memcache stats structure");

  initStringInfo (&buf);

  appendStringInfo (&buf, "pid: %u\n", stats->pid);
  appendStringInfo (&buf, "uptime: %lu\n", (unsigned long) stats->uptime);
  appendStringInfo (&buf, "threads: %u\n", stats->threads);
  appendStringInfo (&buf, "time: %lu\n", (unsigned long) stats->time);
  appendStringInfo (&buf, "pointer_size: %u\n", stats->pointer_size);
  appendStringInfo (&buf, "rusage_user_seconds: %u\n",
		    stats->rusage_user_seconds);
  appendStringInfo (&buf, "rusage_user_microseconds: %u\n",
		    stats->rusage_user_microseconds);
  appendStringInfo (&buf, "rusage_system_seconds: %u\n",
		    stats->rusage_system_seconds);
  appendStringInfo (&buf, "rusage_system_microseconds: %u\n",
		    stats->rusage_system_microseconds);

  appendStringInfo (&buf, "curr_items: %u\n", stats->curr_items);
  appendStringInfo (&buf, "total_items: %u\n", stats->total_items);
  appendStringInfo (&buf, "limit_maxbytes: %" PRIu64 "\n",
		    stats->limit_maxbytes);
  appendStringInfo (&buf, "curr_connections: %u\n", stats->curr_connections);
  appendStringInfo (&buf, "total_connections: %u\n",
		    stats->total_connections);
  appendStringInfo (&buf, "connection_structures: %u\n",
		    stats->connection_structures);
  appendStringInfo (&buf, "bytes: %" PRIu64 "\n", stats->bytes);
  appendStringInfo (&buf, "cmd_get: %" PRIu64 "\n", stats->cmd_get);
  appendStringInfo (&buf, "cmd_set: %" PRIu64 "\n", stats->cmd_set);
  appendStringInfo (&buf, "get_hits: %" PRIu64 "\n", stats->get_hits);
  appendStringInfo (&buf, "get_misses: %" PRIu64 "\n", stats->get_misses);
  appendStringInfo (&buf, "evictions: %" PRIu64 "\n", stats->evictions);
  appendStringInfo (&buf, "bytes_read: %" PRIu64 "\n", stats->bytes_read);
  appendStringInfo (&buf, "bytes_written: %" PRIu64 "\n",
		    stats->bytes_written);
  appendStringInfo (&buf, "version: %s\n", stats->version);

  PG_RETURN_DATUM (DirectFunctionCall1 (textin, CStringGetDatum (buf.data)));
}


Datum
pgmemcached_stat_get_value (PG_FUNCTION_ARGS)
{
  text *stat = PG_GETARG_TEXT_P (0);
  char *stat_str;
  size_t stat_len;
  memcached_stat_st *stats;
  memcached_return rc;
  StringInfoData buf;

  stat_str = DatumGetCString (DirectFunctionCall1 (textout,
						   PointerGetDatum (stat)));
  stat_len = strlen (stat_str);
  if (stat_len == 0)
    ereport (ERROR,
	     (errcode (ERRCODE_INVALID_PARAMETER_VALUE),
	      errmsg ("memcache statistic key cannot be the empty string")));

  stats = memcached_stat (globals.mc, NULL, &rc);
  globals.rc = rc;
  if (stats == NULL)
    elog (ERROR, "unable to create a new memcache stats structure");

  initStringInfo (&buf);

  if (pg_strncasecmp ("pid", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%u", stats->pid);
  else if (pg_strncasecmp ("uptime", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%lu", (unsigned long) stats->uptime);
  else if (pg_strncasecmp ("threads", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%u", stats->threads);
  else if (pg_strncasecmp ("time", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%lu", (unsigned long) stats->time);
  else if (pg_strncasecmp ("pointer_size", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%u", stats->pointer_size);
  else if (pg_strncasecmp ("rusage_user_seconds", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%u", stats->rusage_user_seconds);
  else if (pg_strncasecmp ("rusage_user_microseconds", stat_str, stat_len) ==
	   0)
    appendStringInfo (&buf, "%u", stats->rusage_user_microseconds);
  else if (pg_strncasecmp ("rusage_system_seconds", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%u", stats->rusage_system_seconds);
  else if (pg_strncasecmp ("rusage_system_microseconds", stat_str, stat_len)
	   == 0)
    appendStringInfo (&buf, "%u", stats->rusage_system_microseconds);

  else if (pg_strncasecmp ("curr_items", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%u", stats->curr_items);
  else if (pg_strncasecmp ("total_items", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%u", stats->total_items);
  else if (pg_strncasecmp ("limit_maxbytes", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%" PRIu64, stats->limit_maxbytes);
  else if (pg_strncasecmp ("curr_connections", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%u", stats->curr_connections);
  else if (pg_strncasecmp ("total_connections", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%u", stats->total_connections);
  else if (pg_strncasecmp ("connection_structures", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%u", stats->connection_structures);
  else if (pg_strncasecmp ("bytes", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%" PRIu64, stats->bytes);
  else if (pg_strncasecmp ("cmd_get", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%" PRIu64, stats->cmd_get);
  else if (pg_strncasecmp ("cmd_set", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%" PRIu64, stats->cmd_set);
  else if (pg_strncasecmp ("get_hits", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%" PRIu64, stats->get_hits);
  else if (pg_strncasecmp ("get_misses", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%" PRIu64, stats->get_misses);
  else if (pg_strncasecmp ("evictions", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%" PRIu64, stats->evictions);
  else if (pg_strncasecmp ("bytes_read", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%" PRIu64, stats->bytes_read);
  else if (pg_strncasecmp ("bytes_written", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%" PRIu64, stats->bytes_written);
  else if (pg_strncasecmp ("version", stat_str, stat_len) == 0)
    appendStringInfo (&buf, "%s", stats->version);
  else
    elog (ERROR, "unknown memcache statistic \"%s\"", stat_str);

  PG_RETURN_DATUM (DirectFunctionCall1 (textin, CStringGetDatum (buf.data)));
}
